import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.plugin.filter.PlugInFilter;
import ij.process.ColorProcessor;
import ij.process.ImageProcessor;

import java.awt.image.PixelGrabber;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * @author Xavier Philippeau
 * 
 * MeanShiftFilter Plugin. 
 *
 * Mean Shift Filter (White Gaussian Noise remover)
 */
public class MeanShiftFilter_ implements PlugInFilter {
	
	// Imput image
	private ImagePlus imgOrig = null;

	// Filter Parameters
	private int size=0;
	private double sigma=0;
	
	// Algorithm Parameters
	private static final int MAXLOOP = 100;
	private static final double EPSILON = 0.1;
	private int SAMPLINGSIZE = 0; // computed from size
	private int SUPPORTSIZE = 0;  // computed from sigma
	private double[] WEIGHT = null;
	
	// Feature space structure
	private class Feature {
		int[] feature;
		double density;
	}
	private int[][][] fimage = null;	
	private Map<Integer,Feature> localfspace = null;
	
	/**
	* This method gets a reference to the image to be locally ranked
	* and returns the filter capabilities.
	* @param	arg	calls for the "about" information.
	* @param	imp	This is the image to be processed
	*
	* @return	Returns DOES_8G + DOES_RGB or DONE.
	*/
	public int setup(String arg, ImagePlus imp) {
		
		// about...
		if (arg.equals("about")) {
			showAbout(); 
			return DONE;
		}

		// else...
		if (imp==null) return DONE;

		// original image
		this.imgOrig = imp ;
		
		// Configuration dialog.
		GenericDialog gd = new GenericDialog("Parameters");
		gd.addNumericField("RGB radius (sigma)",20.00,2);
		gd.addNumericField("Sampling size",5,0);
		
		while(true) {
			gd.showDialog();
			if ( gd.wasCanceled() )	return DONE;
			
			this.sigma = (double) gd.getNextNumber();
			this.size = (int) gd.getNextNumber();
			
			if (this.sigma<=0) continue;
			if (this.size<1) continue;
			break;
		}
		gd.dispose();
		
		return DOES_8G+DOES_RGB;
	}

	/** 
	 * Filters use this method to process the image. If the
	 * SUPPORTS_STACKS flag was set, it is called for each slice in
	 * a stack. ImageJ will lock the image before calling
	 * this method and unlock it when the filter is finished. 
	 */
	public void run(ImageProcessor ip) {
		
		// allocation & precomputation
		this.SAMPLINGSIZE = this.size;
		this.SUPPORTSIZE  = (int)(0.5+this.sigma);
		
		this.WEIGHT = new double[ 1+2*SAMPLINGSIZE*SAMPLINGSIZE ];
		for(int d=0;d<WEIGHT.length;d++)
			WEIGHT[d] = 1.0/(1.0+d/(SAMPLINGSIZE));

		this.localfspace = new HashMap<Integer, Feature>();

		// grab pixels from image
		int[] pixels = new int[this.imgOrig.getWidth()*this.imgOrig.getHeight()];
		try {
			PixelGrabber pg = new PixelGrabber(
					this.imgOrig.getImage(),0,0,this.imgOrig.getWidth(),this.imgOrig.getHeight(),
					pixels,0,this.imgOrig.getWidth()
			);
			pg.grabPixels();
		}
		catch (InterruptedException e) {}
		
		// compute feature space
		this.fimage = new int[this.imgOrig.getWidth()][this.imgOrig.getHeight()][];
		for(int y=0;y<this.imgOrig.getHeight();y++) {
			for(int x=0;x<this.imgOrig.getWidth();x++) {
				int pixel = pixels[x+y*this.imgOrig.getWidth()];
				int[] rgb = new int[3];
				if (this.imgOrig.getBitDepth()==8) { // GRAY image
					rgb[0]=pixel&0xff;
					rgb[1]=rgb[0];
					rgb[2]=rgb[0]; 
				} else { // COLOR image
					rgb[0]=(pixel&0xff0000)>>16;
					rgb[1]=(pixel&0xff00)>>8;
					rgb[2]=pixel&0xff;
				}
				this.fimage[x][y] = rgb;
			}
		}
		
		// the output image	
		ImagePlus imgOut = new ImagePlus(this.imgOrig.getTitle()+"_MeanShift", this.imgOrig.getProcessor());
		ImageProcessor output = new ColorProcessor(imgOut.getWidth(),imgOut.getHeight());
		imgOut.show();
		
		// filtering
		for(int y=0;y<this.imgOrig.getHeight();y++) {
			
			for(int x=0;x<this.imgOrig.getWidth();x++) {
				int[] mode = meanshift(x, y);
				output.set(x, y, getRGB32(mode));
			}
			
			// realtime display
			if (y%10==0) {
				imgOut.setProcessor(null,output);
				imgOut.updateAndDraw();				
				IJ.showProgress((double)y/this.imgOrig.getHeight());
				IJ.showStatus("Line: "+y+"/"+this.imgOrig.getHeight());
			}
		}
		
		// all done
		imgOut.setProcessor(null,output);
		imgOut.updateAndDraw();				
		IJ.showProgress(0);
		IJ.showStatus("Done");
	}

	// RGB array to RGB 32bits
	public int getRGB32(int[]rgb) {
		int alpha=255;
		int r = rgb[0] & 0xFF;
		int g = rgb[1] & 0xFF;
		int b = rgb[2] & 0xFF;
		int c = (alpha<<24)+(r<<16)+(g<<8)+b;
		return c;
	}
	
	// About...
	private void showAbout() {
		IJ.showMessage("Mean-Shift Filter...","Mean-Shift Filter by Pseudocode");
	}

	// ------------------------------------------------------------------------------------
	//                               Local Features Sampling
	// ------------------------------------------------------------------------------------

	private int fhash(int[] feat) {
		return feat[0]<<16 | (feat[1]<<8) | feat[2];
	}

	// return the list of the (weighted) features in the local window
	private Collection<Feature> buildfspace(int x, int y, int window) {
		
		int miny = Math.max(y-window,0);
		int maxy = Math.min(y+window,this.imgOrig.getHeight()-1);
		int minx = Math.max(x-window,0);
		int maxx = Math.min(x+window,this.imgOrig.getWidth()-1);

		this.localfspace.clear();

		for(int yk=miny;yk<=maxy;yk++) {
			for(int xk=minx;xk<=maxx;xk++) {
				int[] feature = this.fimage[xk][yk];
				double density = WEIGHT[(xk-x)*(xk-x)+(yk-y)*(yk-y)];
				
				int fid = fhash(feature);
				Feature f = this.localfspace.get(fid);
				if (f==null) {
					f = new Feature();
					f.feature = feature;
					this.localfspace.put(fid, f);
				}
				f.density += density;
			}
		}
		
		return this.localfspace.values();
	}
	
	// ------------------------------------------------------------------------------------
	//                               MEAN SHIFT ALGORITHM
	// ------------------------------------------------------------------------------------
	
	// compute new values of the MeanShiftVector
	private double[] MSVector(double[] vector, Collection<Feature> features) {
		double[] mh = new double[3];
		
		double wsum=0;
		
		int minf0 = (int)Math.max(vector[0]-SUPPORTSIZE,0);
		int maxf0 = (int)Math.min(vector[0]+SUPPORTSIZE,255);
		
		int minf1 = (int)Math.max(vector[1]-SUPPORTSIZE,0);
		int maxf1 = (int)Math.min(vector[1]+SUPPORTSIZE,255);
		
		int minf2 = (int)Math.max(vector[2]-SUPPORTSIZE,0);
		int maxf2 = (int)Math.min(vector[2]+SUPPORTSIZE,255);

		for(Feature f:features) {
			if (f.feature[0]<minf0 || f.feature[0]>maxf0) continue;
			if (f.feature[1]<minf1 || f.feature[1]>maxf1) continue;
			if (f.feature[2]<minf2 || f.feature[2]>maxf2) continue;
			
			double w = f.density; /*1/(1+FeatureDistance(f.feature,vector));*/
			mh[0]+=w*f.feature[0];
			mh[1]+=w*f.feature[1];
			mh[2]+=w*f.feature[2];
			wsum+=w;
		}
		
		if (wsum>0) {
			mh[0]=(mh[0]/wsum)-vector[0];
			mh[1]=(mh[1]/wsum)-vector[1];
			mh[2]=(mh[2]/wsum)-vector[2];
		}
		
		return mh;
	}
	
	// Initialize then move the MeanShiftVector to the nearest mode
	private int[] meanshift(int x,int y) {
		
		// feature of current pixel
		int[] feature = this.fimage[x][y];

		// local feature-space
		Collection<Feature> fspace = buildfspace(x, y, SAMPLINGSIZE);
		
		// declare meanshift variables
		double[] mode = new double[3];
		double[] mh = null;
		double mhnorm=0;
		
		// initialize mode
		mode[0]=feature[0];
		mode[1]=feature[1];
		mode[2]=feature[2];
		
		// compute Mh
		mh = MSVector(mode,fspace);
		
		// compute |mh|^2
		mhnorm = mh[0]*mh[0] + mh[1]*mh[1] + mh[2]*mh[2];
		
		// shift mode until convergence
		int loop=0;
		while( mhnorm>EPSILON && loop<MAXLOOP) {
			
			if (loop>=MAXLOOP/2)
				System.out.println(x+","+y+" ["+loop+"]= "+Arrays.toString(mode)+" mx:"+Arrays.toString(mh)+" "+mhnorm);

			mode[0]+=mh[0];
			mode[1]+=mh[1];
			mode[2]+=mh[2];
			
			mh = MSVector(mode,fspace);
			
			mhnorm = mh[0]*mh[0] + mh[1]*mh[1] + mh[2]*mh[2];

			loop++;
		}

		feature[0]=(int)(0.5+mode[0]);
		feature[1]=(int)(0.5+mode[1]);
		feature[2]=(int)(0.5+mode[2]);
		
		return feature;
	}

}